
#include "chirp_acl.h"
#include "chirp_protocol.h"

#include "username.h"
#include "stringtools.h"
#include "debug.h"

#include <errno.h>
#include <string.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

static int read_only_mode = 0;

void chirp_acl_force_readonly()
{
	read_only_mode = 1;
}

static int is_a_directory( const char *filename )
{
	struct stat info;

	if(stat(filename,&info)==0) {
		if(S_ISDIR(info.st_mode)) {
			return 1;
		} else {
			return 0;
		}
	} else {
		return 0;
	}
}

static void make_acl_name( const char *filename, int get_parent, char *aclname )
{
	sprintf(aclname,"%s/%s",filename,CHIRP_ACL_BASE_NAME);
}

static int do_chirp_acl_check( const char *dirname, const char *subject, int flags )
{
	FILE *aclfile;
	char aclsubject[CHIRP_LINE_MAX];
	int aclflags;
	int access_granted = 0;

	if(read_only_mode) {
		if((flags&(CHIRP_ACL_READ|CHIRP_ACL_LIST)) != flags) {
			errno = EACCES;
			return 0;
		} else {
			return 1;
		}
	}

	aclfile = chirp_acl_open(dirname);
	if(aclfile) {
		while(chirp_acl_read(aclfile,aclsubject,&aclflags)) {
			if(string_match(aclsubject,subject) || !strcmp(aclsubject,"system:anyuser")) {
				if((flags&aclflags) == flags) {
					access_granted = 1;
					break;
				}
			}
		}
		chirp_acl_close(aclfile);
	} else {
		errno = EACCES;
	}

	if(!access_granted) {
		errno = EACCES;
	}

	return access_granted;
}

int chirp_acl_check( const char *filename, const char *subject, int flags )
{
	char temp[CHIRP_PATH_MAX];
	char dirname[CHIRP_PATH_MAX];

	if(!strcmp(string_back(filename,CHIRP_ACL_BASE_LENGTH),CHIRP_ACL_BASE_NAME)) {
		if(flags!=CHIRP_ACL_READ) {
			errno = EACCES;
			return 0;
		}
	}

	sprintf(temp,"%s/..",filename);
	string_collapse_path(temp,dirname,1);

	return do_chirp_acl_check(dirname,subject,flags);
}

int chirp_acl_check_dir( const char *dirname, const char *subject, int flags )
{
	return do_chirp_acl_check(dirname,subject,flags);
}

int chirp_acl_set( const char *dirname, const char *subject, int flags )
{
	char aclname[CHIRP_PATH_MAX];
	char newaclname[CHIRP_PATH_MAX];
	char aclsubject[CHIRP_LINE_MAX];
	int aclflags;
	FILE *aclfile, *newaclfile;
	int result;
	int replaced_acl_entry=0;

	if(!is_a_directory(dirname)) {
		errno = ENOTDIR;
		return -1;
	}

	sprintf(aclname,"%s/%s",dirname,CHIRP_ACL_BASE_NAME);
	sprintf(newaclname,"%s/%s.%d",dirname,CHIRP_ACL_BASE_NAME,getpid());

	aclfile = fopen(aclname,"r");

	/* If the acl never existed, then we can simply create it. */
	if(!aclfile && errno==ENOENT) {
		aclfile = fopen("/dev/null","r");
	}

	if(!aclfile) {
		errno = EACCES;
		return -1;
	}

	replaced_acl_entry = 0;

	newaclfile = fopen(newaclname,"w");
	if(!newaclfile) {
		fclose(aclfile);
		errno = EACCES;
		return -1;
	}

	while(chirp_acl_read(aclfile,aclsubject,&aclflags)) {
		if(!strcmp(subject,aclsubject)) {
			aclflags = flags;
			replaced_acl_entry = 1;
		}
		if(aclflags!=0) {
			fprintf(newaclfile,"%s %s\n",aclsubject,chirp_acl_flags_to_text(aclflags));
		}
	}

	if(!replaced_acl_entry) {
		fprintf(newaclfile,"%s %s\n",subject,chirp_acl_flags_to_text(flags));
	}

	if(ferror(newaclfile)) {
		errno = EACCES;
		result = -1;
	} else {
		result = rename(newaclname,aclname);
		if(result<0) {
			unlink(newaclname);
			errno = EACCES;
			result = -1;
		}
	}

	fclose(aclfile);
	fclose(newaclfile);

	return result;
}


FILE * chirp_acl_open( const char *dirname )
{
	char aclname[CHIRP_PATH_MAX];
	if(!is_a_directory(dirname)) {
		errno = ENOTDIR;
		return 0;
	} else {
		make_acl_name(dirname,0,aclname);
		return fopen(aclname,"r");
	}
}

int chirp_acl_read( FILE *aclfile, char *subject, int *flags )
{
	char acl[CHIRP_LINE_MAX];
	char tmp[CHIRP_LINE_MAX];

	while(fgets(acl,sizeof(acl),aclfile)) {
		if(sscanf(acl,"%[^ ] %[rwlva]",subject,tmp)==2) {
			*flags = chirp_acl_text_to_flags(tmp);
			return 1;
		} else {
			continue;
		}
	}

	return 0;
}

void chirp_acl_close( FILE *aclfile )
{
	fclose(aclfile);
}


const char * chirp_acl_flags_to_text( int flags )
{
	static char text[20];

	text[0] = 0;
	if(flags&CHIRP_ACL_READ)    strcat(text,"r");
	if(flags&CHIRP_ACL_WRITE)   strcat(text,"w");
	if(flags&CHIRP_ACL_LIST)    strcat(text,"l");
	if(flags&CHIRP_ACL_RESERVE) strcat(text,"v");
	if(flags&CHIRP_ACL_ADMIN)   strcat(text,"a");

	return text;
}

int chirp_acl_text_to_flags( const char *t )
{
	int flags=0;

	while(*t) {
		if( *t=='r' ) flags|=CHIRP_ACL_READ;
		if( *t=='w' ) flags|=CHIRP_ACL_WRITE;
		if( *t=='l' ) flags|=CHIRP_ACL_LIST;
		if( *t=='v' ) flags|=CHIRP_ACL_RESERVE;
		if( *t=='a' ) flags|=CHIRP_ACL_ADMIN;
		t++;
	}

	return flags;
}

int chirp_acl_init_root( const char *path )
{
	char aclpath[CHIRP_PATH_MAX];
	char username[USERNAME_MAX];
	FILE *file;

	file = chirp_acl_open(path);
	if(file) {
		chirp_acl_close(file);
		return 1;
	}

	username_get(username);

	sprintf(aclpath,"%s/%s",path,CHIRP_ACL_BASE_NAME);
	file = fopen(aclpath,"w");
	if(file) {
		fprintf(file,"unix:%s %s\n",username,chirp_acl_flags_to_text(CHIRP_ACL_ALL));
		fclose(file);
		return 1;
	} else {
		return 0;
	}
}

int chirp_acl_init_copy( const char *path )
{
	char oldpath[CHIRP_LINE_MAX];
	char newpath[CHIRP_LINE_MAX];
	char subject[CHIRP_LINE_MAX];
	FILE *oldfile;
	FILE *newfile;
	int result = 0;
	int flags;

	sprintf(oldpath,"%s/..",path);
	sprintf(newpath,"%s/%s",path,CHIRP_ACL_BASE_NAME);

	oldfile = chirp_acl_open(oldpath);
	if(oldfile) {
		newfile = fopen(newpath,"w");
		if(newfile) {
			while(chirp_acl_read(oldfile,subject,&flags)) {
				fprintf(newfile,"%s %s\n",subject,chirp_acl_flags_to_text(flags));
			}
			fclose(newfile);
			result = 1;
		}
		chirp_acl_close(oldfile);
	}

	return result;
}

int chirp_acl_init_reserve( const char *path, const char *subject )
{
	char aclpath[CHIRP_PATH_MAX];
	FILE *file;

	sprintf(aclpath,"%s/%s",path,CHIRP_ACL_BASE_NAME);
	file = fopen(aclpath,"w");
	if(file) {
		fprintf(file,"%s %s\n",subject,chirp_acl_flags_to_text(CHIRP_ACL_ALL));
		fclose(file);
		return 1;
	} else {
		return 0;
	}
}
