/*
Copyright (C) 2004 Douglas Thain
This work is made available under a BSD-style license.
See the file COPYING for details.
*/

#include "chirp_resolve.h"
#include "chirp_protocol.h"
#include "chirp_reli.h"

#include "debug.h"
#include "stringtools.h"
#include "xmalloc.h"
#include "hash_table.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

struct mount_entry {
	char prefix[CHIRP_PATH_MAX];
	char redirect[CHIRP_PATH_MAX];
	struct mount_entry *next;
};

struct mount_entry * manual_list = 0;
struct mount_entry * file_list = 0;

static const char *chirp_resolver = 0;
static const char *file_resolver = 0;

static time_t file_resolver_mtime = 0;
static struct hash_table *resolve_cache = 0;

static char * lookup_env( const char *name, void *arg )
{
	char *value = getenv(name);
	if(value) {
		return strdup(value);
	} else {
		fatal("mountfile contains an undefined variable: %s\n",name);
		return 0;
	}
}

static void add_mount_entry( const char *prefix, const char *redirect, struct mount_entry **list )
{
	struct mount_entry * m = xxmalloc(sizeof(*m));
	strcpy(m->prefix,prefix);
	strcpy(m->redirect,redirect);
	m->next = *list;
	*list = m;
}

static void clear_mount_entries( struct mount_entry **list )
{
	struct mount_entry *m;

	while(*list) {
		m = *list;
		*list = m->next;
		free(m);
	}
}

void chirp_resolve_manual_config( const char *str )
{
	char *e = strchr(str,'=');
	if(!e) fatal("badly formed mount string: %s",str);

	*e = 0;
	e++;

	add_mount_entry(str,e,&manual_list);
}

void chirp_resolve_chirp_config( const char *hostname )
{
	chirp_resolver = xstrdup(hostname);
}

void chirp_resolve_file_config( const char *filename )
{
	file_resolver = xstrdup(filename);
}


static void update_file_resolver()
{
	FILE *file;
	char line[CHIRP_LINE_MAX];
	char prefix[CHIRP_LINE_MAX];
	char redirect[CHIRP_LINE_MAX];
	int fields;
	int linenum=0;
	int result;
	struct stat buf;

	result = stat(file_resolver,&buf);
	if(result!=0) fatal("couldn't stat mountfile %s: %s\n",file_resolver,strerror(errno));

	if(buf.st_mtime==file_resolver_mtime) return;
	file_resolver_mtime = buf.st_mtime;

	clear_mount_entries(&file_list);

	file = fopen(file_resolver,"r");
	if(!file) fatal("couldn't open mountfile %s: %s\n",file_resolver,strerror(errno));

	while(1) {
		if(!fgets(line,sizeof(line),file)) {
			if(errno==EINTR) {
				continue;
			} else {
				break;
			}
		}
		linenum++;
		if(line[0]=='#') continue;
		string_chomp(line);
		if(!line[0]) continue;
		fields = sscanf(line,"%s %s",prefix,redirect);

		if(fields==0) {
			continue;
		} else if(fields<2) {
			fatal("%s has an error on line %d\n",file_resolver,linenum);
		} else {
			add_mount_entry(prefix,redirect,&file_list);
		}
	}

	fclose(file);
}

/*
Compare a logical name to a mountlist entry and
determine what to do with it.
*/

static chirp_resolve_t mount_entry_check( const char *logical_name, const char *prefix, const char *redirect, char *physical_name )
{
	chirp_resolve_t result;

	int plen = strlen(prefix);
	int llen = strlen(logical_name);
	if(
		!strncmp(prefix,logical_name,plen) &&
		(
			prefix[plen-1]=='/' ||
			logical_name[plen]=='/' ||
			plen==llen
		)
	) {
		if(!strcmp(redirect,"DENY")) {
			result = CHIRP_RESOLVE_DENIED;
		} else {
			strcpy(physical_name,redirect);
			if(llen>plen) {
				strcat(physical_name,"/");
				strcat(physical_name,&logical_name[plen]);
			}
			result = CHIRP_RESOLVE_CHANGED;
		}
	} else {
		result = CHIRP_RESOLVE_UNCHANGED;
	}

	return result;
}

static chirp_resolve_t resolve_manual( const char *logical_name, char *physical_name  )
{
	struct mount_entry *e;
	chirp_resolve_t result = CHIRP_RESOLVE_UNCHANGED;

	for(e=manual_list;e;e=e->next) {
		result = mount_entry_check(logical_name,e->prefix,e->redirect,physical_name);
		if(result!=CHIRP_RESOLVE_UNCHANGED) break;
	}

	return result;
}

static chirp_resolve_t resolve_file( const char *logical_name, char *physical_name )
{
	struct mount_entry *e;
	chirp_resolve_t result = CHIRP_RESOLVE_UNCHANGED;

	if(!file_resolver) return CHIRP_RESOLVE_UNCHANGED;

	update_file_resolver();

	for(e=file_list;e;e=e->next) {
		result = mount_entry_check(logical_name,e->prefix,e->redirect,physical_name);
		if(result!=CHIRP_RESOLVE_UNCHANGED) break;
	}

	return result;
}

static chirp_resolve_t resolve_chirp( const char *logical_name, char *physical_name, time_t stoptime )
{
	char *temp;
	int length;

	if(!chirp_resolver) return CHIRP_RESOLVE_UNCHANGED;

	length = chirp_reli_lookup(chirp_resolver,logical_name,&temp,stoptime);
	if(length>0) {
		strncpy(physical_name,temp,length);
		physical_name[length] = 0;
		free(temp);
		return CHIRP_RESOLVE_CHANGED;
	} else if(length==CHIRP_ERROR_NOT_AUTHORIZED) {
		return CHIRP_RESOLVE_DENIED;
	} else {
		return CHIRP_RESOLVE_FAILED;
	}
}

/*
Some services, such as the Condor chirp proxy,
will give us unusual url-looking paths like buffer:remote:/biz/foo.
Clean these up into a form that we can use.
*/

void clean_up_path( char *path )
{
	char temp[CHIRP_PATH_MAX];
	char prefix[CHIRP_PATH_MAX];
	char rest[CHIRP_PATH_MAX];

	while(1) {
		if(!strncmp(path,"buffer:",7)) {
			strcpy(temp,path+7);
			strcpy(path,temp);
		} else {
			if(sscanf(path,"%[^:/]:%s",prefix,rest)==2) {
				int plen = strlen(prefix);
				if(!strcmp(prefix,"remote")) {
					strcpy(prefix,"chirp/CONDOR");
				}
				sprintf(temp,"/%s/%s",prefix,path+plen+2);
				strcpy(path,temp);
			} else {
				break;
				/* leave it as is */
			}
		}
	}
}

chirp_resolve_t chirp_resolve( const char *logical_name, char *physical_name, time_t stoptime )
{
	chirp_resolve_t result = CHIRP_RESOLVE_UNCHANGED;
	char *t;

	if(!resolve_cache) resolve_cache = hash_table_create(0,0);

	t = hash_table_lookup(resolve_cache,logical_name);
	if(t) {
		strcpy(physical_name,t);
		result = CHIRP_RESOLVE_CHANGED;
	} else {
		result = resolve_manual(logical_name,physical_name);
		if(result==CHIRP_RESOLVE_UNCHANGED)
			result = resolve_file(logical_name,physical_name);
		if(result==CHIRP_RESOLVE_UNCHANGED)
			result = resolve_chirp(logical_name,physical_name,stoptime);
	}

	switch(result) {
		case CHIRP_RESOLVE_UNCHANGED:
			strcpy(physical_name,logical_name);
			break;
		case CHIRP_RESOLVE_CHANGED:
			clean_up_path(physical_name);
			t = xstrdup(physical_name);
			t = string_subst(t,lookup_env,0);
			strcpy(physical_name,t);
			free(t);
			break;
		case CHIRP_RESOLVE_FAILED:
			debug(D_RESOLVE,"%s failed",logical_name);
			break;
		case CHIRP_RESOLVE_DENIED:
			debug(D_RESOLVE,"%s denied",logical_name);
			break;
	}

	if(result==CHIRP_RESOLVE_UNCHANGED || result==CHIRP_RESOLVE_CHANGED) {
		debug(D_RESOLVE,"%s = %s",logical_name,physical_name);

		if(!hash_table_lookup(resolve_cache,logical_name)) {
			hash_table_insert(resolve_cache,logical_name,xstrdup(physical_name),0);
		}
	}

	return result;
}


