/*
Copyright (C) 2004 Douglas Thain
This work is made available under a BSD-style license.
See the file COPYING for details.
*/

#include "chirp_protocol.h"
#include "chirp_acl.h"

#include "debug.h"
#include "link.h"
#include "getopt.h"
#include "auth_all.h"
#include "stringtools.h"
#include "full_io.h"
#include "get_canonical_path.h"
#include "datagram.h"
#include "username.h"
#include "catalog_server.h"
#include "disk_info.h"
#include "domain_name_cache.h"
#include "create_dir.h"
#include "delete_dir.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/statfs.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <signal.h>
#include <utime.h>
#include <sys/wait.h>
#include <pwd.h>

static void chirp_receive( struct link *l );
static void chirp_handler( struct link *l, const char *subject );
static int errno_to_chirp( int e );
static int access_flags_to_acl( int flags );
static int open_flags_to_acl( int flags );

static int port = CHIRP_PORT;
static int idle_timeout = 86400; /* one day */
static int stall_timeout = 3600; /* one hour */
static int advertise_timeout = 300; /* five minutes */
static int single_mode = 0;
static const char *catalog_host = CATALOG_HOST;
static char owner[USERNAME_MAX] = "";
static time_t starttime;
static struct datagram *catalog_port;
static char hostname[DOMAIN_NAME_MAX];
static char address[LINK_ADDRESS_MAX];
static char rootpath[CHIRP_PATH_MAX];
static const char *startdir = ".";
static const char *safe_username = 0;
static uid_t safe_uid = 0;
static gid_t safe_gid = 0;

static void show_version( const char *cmd )
{
        printf("%s version %d.%d.%d built by %s@%s on %s at %s\n",cmd,CCTOOLS_VERSION_MAJOR,CCTOOLS_VERSION_MINOR,CCTOOLS_VERSION_MICRO,BUILD_USER,BUILD_HOST,__DATE__,__TIME__);
}

static void show_help( const char *cmd )
{
	printf("use: %s [options]\n",cmd);
	printf("where options are:\n");
	printf(" -d <flag>  Enable debugging for this sybsystem\n");
	printf(" -c <dir>   Challenge directory for filesystem authentication.\n");
	printf(" -h         This message.\n");
	printf(" -o <file>  Send debugging output to this file.\n");
	printf(" -O <bytes> Rotate file once it reaches this size.\n");
	printf(" -p <port>  Listen on this port (default is %d)\n",port);
	printf(" -r <dir>   Root of storage directory. (default is current dir)\n");
	printf(" -R         Read-only / read-everything mode.\n");
	printf(" -s <time>  Abort stalled operations after this long. (default is 1h)\n");
	printf(" -S         Single process mode, do not fork.\n");
	printf(" -t <time>  Disconnect idle clients after this time. (default is 1d)\n");
	printf(" -u <host>  Send status updates to this host. (default is %s)\n",CATALOG_HOST);
	printf(" -U <time>  Send status updates at this interval. (default is 5m)\n");
	printf(" -v         Show version info.\n");
	printf(" -w <name>  The name of this server's owner.  (default is username)\n");

	printf("\n");
	printf("Where debug flags are: ");
	debug_flags_print(stdout);
	printf("\n\n");
	printf("Time arguments to -t, -s, and -U are assumed to be in seconds unless followed\n");
	printf("by an 's' 'm' 'h' or 'd' to indicate seconds, minutes, hours, or days.\n");
	printf("For example: chirp_server -t 7d -s 2h\n");
}

static int parse_time_string( const char *str )
{
	int value;
	char mod;

	if(sscanf(str,"%d%c",&value,&mod)==2) {
		switch(mod) {
			case 's': return value;
			case 'm': return value*60;
			case 'h': return value*60*60;
			case 'd': return value*60*60*24;
		}
	} else if(sscanf(str,"%d",&value)==1) {
		return value;
	}

	fprintf(stderr,"Sorry, I don't understand the time value '%s'\n",str);
	exit(1);
}

void shutdown_clean( int sig )
{
	exit(0);
}

void ignore_signal( int sig )
{
}

void reap_child( int sig )
{
	pid_t pid;
	int status;

	do {
		pid = waitpid(-1,&status,WNOHANG);
	} while(pid>0);
}

static void install_handler( int sig, void (*handler)(int sig))
{
	struct sigaction s;
	s.sa_handler = handler;
	sigfillset(&s.sa_mask);
	s.sa_flags = 0; 
	sigaction(sig,&s,0);
}

static void update_catalog()
{
	char text[DATAGRAM_PAYLOAD_MAX];
	char addr[DATAGRAM_ADDRESS_MAX];
	UINT64_T avail, total;
	unsigned uptime;
	int length;

	disk_info_get(".",&avail,&total);
	uptime = time(0)-starttime;

	length = sprintf(
		text,
		"type chirp\nversion %d.%d.%d\nurl chirp://%s:%d\nname %s\nowner %s\ntotal %llu\navail %llu\nuptime %u\nport %d\n",
		CCTOOLS_VERSION_MAJOR,
		CCTOOLS_VERSION_MINOR,
		CCTOOLS_VERSION_MICRO,
		hostname,
		port,
		hostname,
		owner,
		total,
		avail,
		uptime,
		port);

	if(domain_name_cache_lookup(catalog_host,addr)) {
		debug(D_DEBUG,"sending update to %s:%d",catalog_host,CATALOG_PORT);
		datagram_send(catalog_port,text,length,addr,CATALOG_PORT);
	}
}

int main( int argc, char *argv[] )
{
	struct link *link;
	char c;

	debug_config(argv[0]);

	while((c=getopt(argc,argv,"Aa:c:t:i:I:s:Sp:r:Ro:O:d:vw:u:U:hi"))!=(char)-1) {
		switch(c) {
			case 'a':
				fprintf(stderr,"The -a authfile option is no longer used.\n");
				fprintf(stderr,"You should be using Chirp ACLs instead.\n");
				fprintf(stderr,"Please the manual for more details.\n");
				exit(1);
				break;
			case 'c':
				auth_unix_challenge_dir(optarg);
				break;
			case 'd':
				debug_flags_set(optarg);
				break;
			case 'i':
				safe_username = optarg;
				break;
			case 'p':
				port = atoi(optarg);
				break;
			case 't':
				idle_timeout = parse_time_string(optarg);
				break;
			case 's':
				stall_timeout = parse_time_string(optarg);
				break;
			case 'S':
				single_mode = 1;
				break;
			case 'r':
				startdir = optarg;
				break;
			case 'R':
				chirp_acl_force_readonly();
				break;
			case 'o':
				debug_config_file(optarg);
				break;
			case 'O':
				debug_config_file_size(atoi(optarg));
				break;
			case 'u':
				catalog_host = optarg;
				break;
			case 'U':
				advertise_timeout = parse_time_string(optarg);
				break;
			case 'v':
				show_version(argv[0]);
				return 1;
			case 'w':
				strcpy(owner,optarg);
				break;
			case 'h':
			default:
				show_help(argv[0]);
				return 1;
		}
	}

	if(getuid()==0) {
		struct passwd *p;
		if(!safe_username) {		
			printf("Sorry, I refuse to run as root without certain safeguards.\n");
			printf("Please give me a safe username with the -i <user> option.\n");
			printf("After using root access to authenticate users,\n");
			printf("I will use the save username to access data on disk.\n");
			exit(1);
		}
		p = getpwnam(safe_username);
		if(!p) fatal("unknown user: %s",safe_username);
		safe_uid = p->pw_uid;
		safe_gid = p->pw_gid;
		chown(startdir,safe_uid,safe_gid);
		chmod(startdir,0700);
	} else {
		if(safe_username) {
			printf("Sorry, the -i option doesn't make sense\n");
			printf("unless I am already running as root.\n");
			exit(1);
		}
	}

	if(!create_dir(startdir,0700))
		fatal("couldn't create %s: %s\n",startdir,strerror(errno));
	if(!chirp_acl_init_root(startdir))
		fatal("couldn't create root acl in %s: %s\n",startdir,strerror(errno));
	if(chdir(startdir)!=0)
		fatal("coudln't move to %s: %s\n",startdir,strerror(errno));
	if(getcwd(rootpath,sizeof(rootpath))<0)
		fatal("couldn't get working dir: %s\n",strerror(errno));

	link = link_serve(port);
	if(!link) fatal("couldn't listen on port %d: %s",port,strerror(errno));

	link_address_local(link,address,&port);

	auth_register_all();
	starttime = time(0);
	catalog_port = datagram_create(0);
	domain_name_cache_guess(hostname);
	if(!owner[0]) {
		if(!username_get(owner)) {
			strcpy(owner,"unknown");
		}
	}

	install_handler(SIGPIPE,ignore_signal);
	install_handler(SIGHUP,ignore_signal);
	install_handler(SIGCHLD,reap_child);
	install_handler(SIGINT,shutdown_clean);
	install_handler(SIGTERM,shutdown_clean);
	install_handler(SIGQUIT,shutdown_clean);

	while(1) {
		struct link *l;
		pid_t pid;

		update_catalog();

		l = link_accept(link,time(0)+advertise_timeout);
		if(!l) continue;

		if(single_mode) {
			chirp_receive(l);
		} else {
			pid = fork();
			if(pid==0) {
				/* turn off the automatic waitpid for grandchildren */
				install_handler(SIGCHLD,ignore_signal);
				chirp_receive(l);
				_exit(0);
			} else if(pid>=0) {
				debug(D_CHIRP,"created pid %d",pid);
			} else {
				debug(D_CHIRP,"couldn't fork: %s",strerror(errno));
			}
			link_close(l);
		}
	}
}

static void chirp_receive( struct link *link )
{
	char *atype, *asubject;
	char typesubject[AUTH_TYPE_MAX+AUTH_SUBJECT_MAX];
	char addr[LINK_ADDRESS_MAX];
	int port;

	link_address_remote(link,addr,&port);
	debug(D_AUTH,"client connected from %s:%d",addr,port);

	if(auth_accept(link,&atype,&asubject,time(0)+stall_timeout)) {
		sprintf(typesubject,"%s:%s",atype,asubject);
		free(atype);
		free(asubject);
		debug(D_AUTH,"client authenticates as %s",typesubject);
		if(safe_username) {
			debug(D_AUTH,"changing to uid %d gid %d",safe_uid,safe_gid);
			setgid(safe_gid);
			setuid(safe_uid);
		}			
		chirp_handler(link,typesubject);
	} else {
		debug(D_AUTH,"client could not authenticate");
	}

	link_close(link);
}

/*
Force a path to fall within the simulated root directory.
*/

static int chirp_path_fix( char *path )
{
	char temp[CHIRP_PATH_MAX*2];
	int rlen;

	// Put the root path into place:
	rlen = sprintf(temp,"%s/",rootpath);

	// Now append the canonicalized path with dots and such removed.
	string_collapse_path(path,&temp[rlen],1);

	// Copy the fixed path back out.
	strcpy(path,temp);

	return 1;
}


static void chirp_handler( struct link *l, const char *subject )
{
	char line[CHIRP_LINE_MAX];
	
	link_tune(l,LINK_TUNE_INTERACTIVE);

	while(1) {
		int do_stat_result = 0;
		int do_statfs_result = 0;
		int do_getdir_result = 0;
		int do_no_result = 0;
		int result = -1;

		char *dataout = 0;
		int dataoutlength = 0;

		char path[CHIRP_PATH_MAX];
		char newpath[CHIRP_PATH_MAX];
		char newsubject[CHIRP_LINE_MAX];
		char newacl[CHIRP_LINE_MAX];
		int fd, offset, length, flags;
		int actual;
		uid_t uid;
		gid_t gid;
		mode_t mode;
		struct stat statbuf;
		struct statfs statfsbuf;
		struct utimbuf ut;
		time_t idletime = time(0)+idle_timeout;
		time_t stalltime;

		if(!link_readline(l,line,sizeof(line),idletime)) {
			if(time(0)>=idletime) {
				debug(D_CHIRP,"timeout: client idle too long\n");
				sprintf(line,"%d\n",CHIRP_ERROR_TIMED_OUT);
				link_write(l,line,strlen(line),time(0)+stall_timeout);
			}
			break;
		}

		stalltime = time(0)+stall_timeout;
		string_chomp(line);
		if(strlen(line)<1) continue;
		if(line[0]==4) break;

		debug(D_CHIRP,"%s",line);

		if(sscanf(line,"pread %d %d %d",&fd,&length,&offset)==3) {
			dataout = malloc(length);
			if(dataout) {
				result = full_pread(fd,dataout,length,offset);
				if(result>=0) {
					dataoutlength = result;
				} else {
					free(dataout);
					dataout = 0;
				}
			} else {
				errno = ENOMEM;
			}
		} else if(sscanf(line,"whoami %d",&length)==1) {
			if(strlen(subject)<length) length = strlen(subject);
			dataout = malloc(length);
			if(dataout) {
				dataoutlength = length;
				strncpy(dataout,subject,length);
				result = length;
			} else {
				result = -1;
			}
		} else if(sscanf(line,"readlink %s %d",path,&length)==2) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_READ)) goto failure;
			dataout = malloc(length);
			if(dataout) {
				result = readlink(path,dataout,length);
				if(result>=0) {
					/* Careful: readlink could expose the local root path. */
					/* If it matches, remove the prefix, otherwise panic. */
					int rlen = strlen(rootpath);
					if(!strncmp(rootpath,dataout,rlen)) {
						strcpy(dataout,&dataout[rlen]);
						result = result-rlen;
						dataoutlength = result;
					} else {
						free(dataout);
						dataout = 0;
						errno = EINVAL;
						result = -1;
					}
				} else {
					free(dataout);
					dataout = 0;
				}
			} else {
				errno = ENOMEM;
				result = -1;
			}
		} else if(sscanf(line,"pwrite %d %d %d",&fd,&length,&offset)==3) {
			char *data = malloc(length);
			if(data) {
				actual = link_read(l,data,length,stalltime);
				if(actual!=length) break;

				result = full_pwrite(fd,data,length,offset);
				free(data);
			} else {
				/* we ought to clear the stream and continue here */
				break;
			}
		} else if(sscanf(line,"getdir %s",path)==1) {
			DIR *dir;
			struct dirent *d;

			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check_dir(path,subject,CHIRP_ACL_LIST)) goto failure;

			dir = opendir(path);
			if(dir) {
				link_write(l,"0\n",2,stalltime);
				while((d=readdir(dir))) {
					if(!strcmp(d->d_name,CHIRP_ACL_BASE_NAME)) continue;
					sprintf(line,"%s\n",d->d_name);
					link_write(l,line,strlen(line),stalltime);
				}
				closedir(dir);
				do_getdir_result = 1;
				result = 0;
			} else {
				result = -1;
			}
		} else if(sscanf(line,"getacl %s",path)==1) {
			char aclsubject[CHIRP_LINE_MAX];
			int aclflags;
			FILE *aclfile;

			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check_dir(path,subject,CHIRP_ACL_LIST)) goto failure;

			aclfile = chirp_acl_open(path);
			if(aclfile) {
				link_write(l,"0\n",2,stalltime);
				while(chirp_acl_read(aclfile,aclsubject,&aclflags)) {
					sprintf(line,"%s %s\n",aclsubject,chirp_acl_flags_to_text(aclflags));
					link_write(l,line,strlen(line),stalltime);
				}
				chirp_acl_close(aclfile);
				do_getdir_result = 1;
				result = 0;
			} else {
				result = -1;
			}
		} else if(sscanf(line,"getfile %s",path)) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_READ)) goto failure;

			fd = open(path,O_RDONLY);
			if(fd>=0) {
				length = lseek(fd,0,SEEK_END);
				lseek(fd,0,SEEK_SET);
				sprintf(line,"%d\n",length);
				link_write(l,line,strlen(line),stalltime);
				result = link_stream_from_fd(l,fd,length,stalltime);
				close(fd);
				if(result!=length) break;
				do_no_result = 1;
			} else {
				result = -1;
			}
		} else if(sscanf(line,"putfile %s %d %d",path,&mode,&length)) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_WRITE)) goto failure;

			/* remove the file to overcome permissions issues */
			unlink(path);

			fd = open(path,O_WRONLY|O_CREAT|O_TRUNC,mode);
			if(fd>=0) {
				sprintf(line,"0\n");
				link_write(l,line,strlen(line),stalltime);
				actual = link_stream_to_fd(l,fd,length,stalltime);
				if(actual!=length) {
					if(actual>=0) link_soak(l,length-actual,stalltime);
					result = -1;
				} else {
					result = actual;
				}
				close(fd);
			} else {
				result = -1;
			}
		} else if(sscanf(line,"open %s %s %d",path,newpath,&mode)==3) {
			flags = 0;

			if(strchr(newpath,'r')) {
				if(strchr(newpath,'w')) {
					flags = O_RDWR;
				} else {
					flags = O_RDONLY;
				}
			} else if(strchr(newpath,'w')) {
				flags = O_WRONLY;
			}

			if(strchr(newpath,'c')) flags |= O_CREAT;
			if(strchr(newpath,'t')) flags |= O_TRUNC;
			if(strchr(newpath,'a'))	flags |= O_APPEND;
			if(strchr(newpath,'x')) flags |= O_EXCL;
			if(strchr(newpath,'s')) flags |= O_SYNC;

			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,open_flags_to_acl(flags))) goto failure;

			result = open(path,flags,mode);
			if(result>=0) {
				fstat(result,&statbuf);
				do_stat_result = 1;
			}

		} else if(sscanf(line,"close %d",&fd)==1) {
			result = close(fd);
		} else if(sscanf(line,"fchmod %d %d",&fd,&mode)==2) {
			result = fchmod(fd,mode);			
		} else if(sscanf(line,"fchown %d %d %d",&fd,&uid,&gid)==3) {
			result = 0;
		} else if(sscanf(line,"fsync %d",&fd)==1) {
			result = fsync(fd);
		} else if(sscanf(line,"ftruncate %d %d",&fd,&length)==2) {
			result = ftruncate(fd,length);
		} else if(sscanf(line,"unlink %s",path)==1) {
			if(!chirp_path_fix(path)) goto failure;
			if(
			   chirp_acl_check(path,subject,CHIRP_ACL_WRITE) ||
			   chirp_acl_check_dir(path,subject,CHIRP_ACL_WRITE)
			) {
				result = unlink(path);
			} else {
				goto failure;
			}
		} else if(sscanf(line,"access %s %d",path,&flags)==2) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,access_flags_to_acl(flags))) goto failure;
			result = access(path,flags);
		} else if(sscanf(line,"chmod %s %d",path,&mode)==2) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_WRITE)) goto failure;
			result = chmod(path,mode);
		} else if(sscanf(line,"chown %s %d %d",path,&uid,&gid)==3) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_WRITE)) goto failure;
			result = 0;
		} else if(sscanf(line,"lchown %s %d %d",path,&uid,&gid)==3) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_WRITE)) goto failure;
			result = 0;
		} else if(sscanf(line,"truncate %s %d",path,&length)==2) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_WRITE)) goto failure;
			result = truncate(path,length);
		} else if(sscanf(line,"rename %s %s",path,newpath)==2) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_path_fix(newpath)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_READ|CHIRP_ACL_WRITE)) goto failure;
			if(!chirp_acl_check(newpath,subject,CHIRP_ACL_READ|CHIRP_ACL_WRITE)) goto failure;
			result = rename(path,newpath);
		} else if(sscanf(line,"link %s %s",path,newpath)==2) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_path_fix(newpath)) goto failure;
			if(!chirp_acl_check(newpath,subject,CHIRP_ACL_WRITE)) goto failure;
			result = link(path,newpath);
		} else if(sscanf(line,"symlink %s %s",path,newpath)==2) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_path_fix(newpath)) goto failure;
			if(!chirp_acl_check(newpath,subject,CHIRP_ACL_WRITE)) goto failure;
			result = symlink(path,newpath);
		} else if(sscanf(line,"setacl %s %s %s",path,newsubject,newacl)==3) {
			if(!chirp_path_fix(path)) goto failure;
			if(
			    chirp_acl_check(path,subject,CHIRP_ACL_ADMIN)  ||
			    chirp_acl_check_dir(path,subject,CHIRP_ACL_ADMIN) 
			) {
				result = chirp_acl_set(path,newsubject,chirp_acl_text_to_flags(newacl));
			} else {
				goto failure;
			}
		} else if(sscanf(line,"mkdir %s %d",path,&mode)==2) {
			if(!chirp_path_fix(path)) goto failure;
			if(chirp_acl_check(path,subject,CHIRP_ACL_WRITE)) {
				result = mkdir(path,mode);
				if(result==0){
					if(chirp_acl_init_copy(path)) {
						result = 0;
					} else {
						rmdir(path);
						errno = EACCES;
						result = -1;
					}
				}
			} else if(chirp_acl_check(path,subject,CHIRP_ACL_RESERVE)) {
				result = mkdir(path,mode);
				if(result==0){
					if(chirp_acl_init_reserve(path,subject)) {
						result = 0;
					} else {
						rmdir(path);
						errno = EACCES;
						result = -1;
					}
				}
			} else {
				goto failure;
			}
		} else if(sscanf(line,"rmdir %s",path)==1) {
			if(!chirp_path_fix(path)) goto failure;
			if(
			    chirp_acl_check(path,subject,CHIRP_ACL_WRITE) ||
			    chirp_acl_check_dir(path,subject,CHIRP_ACL_WRITE)
			) {
				if(delete_dir(path)) {
					result = 0;
				} else {
					result = -1;
				}
			} else {
				goto failure;
			}
		} else if(sscanf(line,"utime %s %lu %lu",path,&ut.actime,&ut.modtime)==3) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_READ)) goto failure;
			result = utime(path,&ut);
		} else if(sscanf(line,"fstat %d",&fd)==1) {
			result = fstat(fd,&statbuf);
			do_stat_result = 1;
		} else if(sscanf(line,"fstatfs %d",&fd)==1) {
#ifdef sun
			result = fstatfs(fd,&statfsbuf,0,0);
#else
			result = fstatfs(fd,&statfsbuf);
#endif
			do_statfs_result = 1;
		} else if(sscanf(line,"statfs %s",path)==1) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_LIST)) goto failure;
#ifdef sun
			result = statfs(path,&statfsbuf,0,0);
#else
			result = statfs(path,&statfsbuf);
#endif
			do_statfs_result = 1;
		} else if(sscanf(line,"stat %s",path)==1) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_LIST)) goto failure;
			result = stat(path,&statbuf);
			do_stat_result = 1;
		} else if(sscanf(line,"lstat %s",path)==1) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_LIST)) goto failure;
			result = lstat(path,&statbuf);
			do_stat_result = 1;
		} else {
			result = -1;
			errno = ENOSYS;
		}

		if(do_no_result) {
			/* nothing */
		} else if(result<0) {
			failure:
			result = errno_to_chirp(errno);
			sprintf(line,"%d\n",result);
		} else if(do_stat_result) {
			sprintf(line,"%lld\n%lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld\n",
				(long long) result,
				(long long) statbuf.st_dev,
				(long long) statbuf.st_ino,
				(long long) statbuf.st_mode,
				(long long) statbuf.st_nlink,
				(long long) statbuf.st_uid,
				(long long) statbuf.st_gid,
				(long long) statbuf.st_rdev,
				(long long) statbuf.st_size,
				(long long) statbuf.st_blksize,
				(long long) statbuf.st_blocks,
				(long long) statbuf.st_atime,
				(long long) statbuf.st_mtime,
				(long long) statbuf.st_ctime);
		} else if(do_statfs_result) {
			sprintf(line,"%lld\n%lld %lld %lld %lld %lld %lld %lld\n",
				(long long) result,
#ifdef sun
				(long long) statfsbuf.f_fstyp,
#else
				(long long) statfsbuf.f_type,
#endif
       				(long long) statfsbuf.f_bsize,
				(long long) statfsbuf.f_blocks,
				(long long) statfsbuf.f_bfree,
#ifdef sun
				(long long) statfsbuf.f_bfree,
#else
				(long long) statfsbuf.f_bavail,
#endif
				(long long) statfsbuf.f_files,
				(long long) statfsbuf.f_ffree);
		} else if(do_getdir_result) {
			sprintf(line,"\n");
		} else {
			sprintf(line,"%d\n",result);
		}

		debug(D_CHIRP,"= %s",line);
		if(!do_no_result) {
			length = strlen(line);
			actual = link_write(l,line,length,stalltime);
			if(actual!=length) break;
		}

		if(dataout) {
			actual = link_write(l,dataout,dataoutlength,stalltime);
			if(actual!=dataoutlength) break;
			free(dataout);
			dataout = 0;
		}

	}
}

static int access_flags_to_acl( int flags )
{
	int acl=0;
	if(flags&R_OK) acl|=CHIRP_ACL_READ;
	if(flags&W_OK) acl|=CHIRP_ACL_WRITE;
	if(flags&X_OK) acl|=CHIRP_ACL_READ;
	return acl;
}

static int open_flags_to_acl( int flags )
{
	int acl = 0;
	if(flags&O_WRONLY) acl|=CHIRP_ACL_WRITE;
	if(flags&O_RDWR)   acl|=CHIRP_ACL_READ|CHIRP_ACL_WRITE;
	if(flags&O_CREAT)  acl|=CHIRP_ACL_WRITE;
	if(flags&O_TRUNC)  acl|=CHIRP_ACL_WRITE;
	if(flags&O_APPEND) acl|=CHIRP_ACL_WRITE;
	if(acl==0) acl=CHIRP_ACL_READ;
	return acl;
}

static int errno_to_chirp( int e )
{
	switch(e) {
		case EACCES:
		case EPERM:
			return CHIRP_ERROR_NOT_AUTHORIZED;
		case ENOENT:
			return CHIRP_ERROR_DOESNT_EXIST;
		case EEXIST:
			return CHIRP_ERROR_ALREADY_EXISTS;
		case EFBIG:
			return CHIRP_ERROR_TOO_BIG;
		case ENOSPC:
			return CHIRP_ERROR_NO_SPACE;
		case ENOMEM:
			return CHIRP_ERROR_NO_MEMORY;
		case ENOSYS:
		case EINVAL:
			return CHIRP_ERROR_INVALID_REQUEST;
		case EMFILE:
		case ENFILE:
			return CHIRP_ERROR_TOO_MANY_OPEN;
		case EBUSY:
			return CHIRP_ERROR_BUSY;
		case EAGAIN:
			return CHIRP_ERROR_TRY_AGAIN;
		case EBADF:
			return CHIRP_ERROR_BAD_FD;
		case EISDIR:
			return CHIRP_ERROR_IS_DIR;
		case ENOTDIR:
			return CHIRP_ERROR_NOT_DIR;
		case ENOTEMPTY:
			return CHIRP_ERROR_NOT_EMPTY;
		case EXDEV:
			return CHIRP_ERROR_CROSS_DEVICE_LINK;
		default:
			debug(D_CHIRP,"zoiks, I don't know how to transform error %d (%s)\n",errno,strerror(errno));
			return CHIRP_ERROR_UNKNOWN;
	}
}
