/*
Copyright (C) 2004 Douglas Thain
This work is made available under a BSD-style license.
See the file COPYING for details.
*/

#include "catalog_server.h"
#include "datagram.h"
#include "link.h"
#include "hash_cache.h"
#include "debug.h"
#include "getopt.h"
#include "nvpair.h"
#include "stringtools.h"
#include "domain_name_cache.h"
#include "username.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <alloca.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>
#include <sys/wait.h>

#ifndef LINE_MAX
#define LINE_MAX 1024
#endif

/* The table of nvpairs, hashed on address:port */
static struct hash_cache *table=0;

/* The time for which updated data lives before automatic deletion */
static int lifetime = 900;

/* The port upon which to listen. */
static int port = CATALOG_PORT;

/* This machine's canonical name. */
static char hostname[DOMAIN_NAME_MAX];

/* This process's owner */
static char owner[USERNAME_MAX];

/* Time when the process was started. */
static time_t starttime;

/* Settings for the master catalog that we will report *to* */
static int outgoing_alarm = 0;
static int outgoing_timeout = 300;
static int outgoing_port = CATALOG_PORT;
static const char *outgoing_host = CATALOG_HOST;

void shutdown_clean( int sig )
{
	exit(0);
}

void ignore_signal( int sig )
{
}

void reap_child( int sig )
{
	pid_t pid;
	int status;

	do {
		pid = waitpid(-1,&status,WNOHANG);
	} while(pid>0);
}

static void install_handler( int sig, void (*handler)(int sig))
{
	struct sigaction s;
	s.sa_handler = handler;
	sigfillset(&s.sa_mask);
	s.sa_flags = 0; 
	sigaction(sig,&s,0);
}

static void update_catalog( struct datagram *outgoing_dgram )
{
	char text[DATAGRAM_PAYLOAD_MAX];
	char addr[DATAGRAM_ADDRESS_MAX];
	unsigned uptime;
	int length;

	uptime = time(0)-starttime;

	length = sprintf(
		text,
		"type catalog\nversion %d.%d.%d\nurl http://%s:%d\nname %s\nowner %s\nuptime %u\nport %d\n",
		CCTOOLS_VERSION_MAJOR,
		CCTOOLS_VERSION_MINOR,
		CCTOOLS_VERSION_MICRO,
		hostname,
		port,
		hostname,
		owner,
		uptime,
		port);

	if(domain_name_cache_lookup(outgoing_host,addr)) {
		debug(D_DEBUG,"sending update to %s:%d",outgoing_host,outgoing_port);
		datagram_send(outgoing_dgram,text,length,addr,outgoing_port);
	}
}

static void handle_updates( struct datagram *update_port )
{
	char data[DATAGRAM_PAYLOAD_MAX*2];
	char addr[DATAGRAM_ADDRESS_MAX];
	char key[DATAGRAM_ADDRESS_MAX*3+3];
	int port;
	int result;
	struct nvpair *nv;
	const char *name, *type;

	while(1) {
		result = datagram_recv(update_port,data,DATAGRAM_PAYLOAD_MAX,addr,&port,0);
		if(result<=0) return;

		data[result]=0;

		nv = nvpair_create();
		nvpair_parse(nv,data);
		nvpair_insert_string(nv,"address",addr);
		nvpair_insert_integer(nv,"lastheardfrom",time(0));

		name = nvpair_lookup_string(nv,"name");
		if(!name) name = "???";

		type = nvpair_lookup_string(nv,"type");
		if(!type) type = "???";

		port = nvpair_lookup_integer(nv,"port");

		sprintf(key,"%s:%s:%s:%d",addr,name,type,port);
		hash_cache_insert(table,key,nv,lifetime);

		debug(D_DEBUG,"received udp update from %s",key);
	}
}

static INT64_T sum_total = 0;
static INT64_T sum_avail = 0;
static INT64_T sum_devices = 0;

static int add_to_totals( struct hash_cache *table, char *key, void *nv, void *stream )
{
	sum_total += nvpair_lookup_integer(nv,"total");
	sum_avail += nvpair_lookup_integer(nv,"avail");
	sum_devices++;
	return 1;
}

static int send_text( struct hash_cache *table, char *key, void *nv, void *stream )
{
	nvpair_print_text(nv,stream);
	return 1;
}

static int send_xml( struct hash_cache *table, char *key, void *nv, void *stream )
{
	nvpair_print_xml(nv,stream);
	return 1;
}

static int send_old_classads( struct hash_cache *table, char *key, void *nv, void *stream )
{
	nvpair_print_old_classads(nv,stream);
	return 1;
}

static int send_new_classads( struct hash_cache *table, char *key, void *nv, void *stream )
{
	nvpair_print_new_classads(nv,stream);
	return 1;
}

static const char *html_headers[] = {
	"type", "name", "owner", "total", "avail", "version", "url", 0,
};

static int send_html( struct hash_cache *table, char *key, void *nv, void *stream )
{
	nvpair_print_html(nv,stream,html_headers);
	return 1;
}

static void handle_query( struct link *query_link )
{
	FILE *stream;
	char line[LINE_MAX];
	char action[LINE_MAX];
	char path[LINE_MAX];
	char version[LINE_MAX];
	char addr[LINK_ADDRESS_MAX];
	int port;
	time_t current;

	link_address_remote(query_link,addr,&port);

	debug(D_DEBUG,"www query from %s:%d",addr,port);

	link_nonblocking(query_link,0);
	stream = fdopen(link_fd(query_link),"r+");
	if(!stream) return;

	if(!fgets(line,sizeof(line),stream)) return;
	string_chomp(line);
	if(sscanf(line,"%s %s %s",action,path,version)!=3) return;

	while(1) {
		if(!fgets(line,sizeof(line),stream)) return;
		if(line[0]=='\n' || line[0]=='\r') break;
	}

	current = time(0);
	fprintf(stream,"HTTP/1.1 200 OK\n");
	fprintf(stream,"Date: %s",ctime(&current));
	fprintf(stream,"Server: catalog_server\n");
	fprintf(stream,"Connection: close\n");
	fprintf(stream,"Content-type: text/");

	if(!strcmp(path,"/query.text")) {
		fprintf(stream,"plain\n\n");
		hash_cache_iterate(table,send_text,stream);
	} else if(!strcmp(path,"/query.oldclassads")) {
		fprintf(stream,"plain\n\n");
		hash_cache_iterate(table,send_old_classads,stream);
	} else if(!strcmp(path,"/query.newclassads")) {
		fprintf(stream,"plain\n\n");
		hash_cache_iterate(table,send_new_classads,stream);
	} else if(!strcmp(path,"/query.xml")) {
		fprintf(stream,"xml\n\n");
		fprintf(stream,"<?xml version=\"1.0\" standalone=\"yes\"?>\n");
		fprintf(stream,"<catalog>\n");
		hash_cache_iterate(table,send_xml,stream);
		fprintf(stream,"</catalog>\n");
	} else {
		char avail_line[LINE_MAX];
		char total_line[LINE_MAX];

		fprintf(stream,"html\n\n");
		fprintf(stream,"<title>%s storage catalog</title>\n",hostname);
		fprintf(stream,"<center>\n");
		fprintf(stream,"<h1>%s storage catalog</h1>\n",hostname);
		fprintf(stream,"<a href=/query.text>text</a> - ");
		fprintf(stream,"<a href=/query.html>html</a> - ");
		fprintf(stream,"<a href=/query.xml>xml</a> - ");
		fprintf(stream,"<a href=/query.oldclassads>oldclassads</a> - ");		
		fprintf(stream,"<a href=/query.newclassads>newclassads</a>");		
		fprintf(stream,"<p>\n");

		sum_total = 0;
		sum_avail = 0;
		sum_devices = 0;
		hash_cache_iterate(table,add_to_totals,0);

		string_metric(sum_avail,-1,avail_line);
		string_metric(sum_total,-1,total_line);
		fprintf(stream,"<b>%sB available out of %sB on %lld devices</b><p>\n",
			avail_line,total_line,sum_devices);

		nvpair_print_html_header(stream,html_headers);
		hash_cache_iterate(table,send_html,stream);
		nvpair_print_html_footer(stream,html_headers);
		fprintf(stream,"</center>\n");
	}

	fclose(stream);
}

static void show_version( const char *cmd )
{
        printf("%s version %d.%d.%d built by %s@%s on %s at %s\n",cmd,CCTOOLS_VERSION_MAJOR,CCTOOLS_VERSION_MINOR,CCTOOLS_VERSION_MICRO,BUILD_USER,BUILD_HOST,__DATE__,__TIME__);
}

static void show_help( const char *cmd )
{
	printf("Use: %s [options]\n",cmd);
	printf("where options are:\n");
	printf(" -p <port>      Port number to listen on (default is %d)\n",port);
	printf(" -l <secs>      Lifetime of data, in seconds (default is %d)\n",lifetime);
	printf(" -d <subsystem> Enable debugging for this subsystem\n");
	printf(" -u <host>      Send status updates to this host. (default is %s)\n",CATALOG_HOST);
	printf(" -U <time>      Send status updates at this interval. (default is 5m)\n");
	printf(" -v             Show version string\n");
	printf(" -h             Show this help screen\n");
}

static int parse_time_string( const char *str )
{
	int value;
	char mod;

	if(sscanf(str,"%d%c",&value,&mod)==2) {
		switch(mod) {
			case 's': return value;
			case 'm': return value*60;
			case 'h': return value*60*60;
			case 'd': return value*60*60*24;
		}
	} else if(sscanf(str,"%d",&value)==1) {
		return value;
	}

	fprintf(stderr,"Sorry, I don't understand the time value '%s'\n",str);
	exit(1);
}

int main( int argc, char *argv[] )
{
	struct datagram *update_dgram=0;
	struct datagram *outgoing_dgram=0;
	struct link *link, *list_port=0;
	char ch;

	debug_config(argv[0]);

	while((ch=getopt(argc,argv,"p:l:d:u:U:hv"))!=(char)-1) {
		switch(ch) {
			case 'd':
				debug_flags_set(optarg);
				break;
			case 'p':
				port = atoi(optarg);
				break;
			case 'u':
				outgoing_host = optarg;
				break;
			case 'U':
				outgoing_timeout = parse_time_string(optarg);
				break;
			case 'l':
				lifetime = parse_time_string(optarg);
				break;
			case 'v':
				show_version(argv[0]);
				return 0;
			case 'h':
			default:
				show_help(argv[0]);
				return 1;
		}
	}

	install_handler(SIGPIPE,ignore_signal);
	install_handler(SIGHUP,ignore_signal);
	install_handler(SIGCHLD,reap_child);
	install_handler(SIGINT,shutdown_clean);
	install_handler(SIGTERM,shutdown_clean);
	install_handler(SIGQUIT,shutdown_clean);

	domain_name_cache_guess(hostname);
	username_get(owner);
	starttime = time(0);

	table = hash_cache_create(127,hash_string,(hash_cache_cleanup_t)nvpair_delete);
	if(!table) fatal("couldn't make hash table");
	
	update_dgram = datagram_create( port );
	if(!update_dgram) fatal("couldn't listen on udp port %d",port);

	outgoing_dgram = datagram_create(0);
	if(!outgoing_dgram) fatal("couldn't create outgoing udp port");

	list_port = link_serve( port );
	if(!list_port) fatal("couldn't listen on tcp port %d",port);

	while(1) {
		handle_updates(update_dgram);

		if(time(0)>outgoing_alarm) {
			update_catalog(outgoing_dgram);
			outgoing_alarm = time(0)+outgoing_timeout;
		}

		link = link_accept(list_port,time(0)+5);
		if(link) {
			pid_t pid = fork();
			if(pid==0) {
				handle_query(link);
				exit(0);
			}
			link_close(link);
		}
	}

	return 1;
}


