/*
Copyright (C) 2004 Douglas Thain
This work is made available under a BSD-style license.
See the file COPYING for details.
*/

#include "hash_cache.h"
#include "hash_table.h"

#include <time.h>
#include <stdlib.h>

struct hash_cache {
	struct hash_table *table;
	hash_cache_cleanup_t cleanup;
};

struct entry {
	void *value;
	time_t expires;
};

struct hash_cache * hash_cache_create( int size, hash_func_t func, hash_cache_cleanup_t cleanup )
{
	struct hash_cache *cache;

	cache = malloc(sizeof(*cache));
	if(!cache) return 0;

	cache->table = hash_table_create(size,func);
	if(!cache->table) {
		free(cache);
		return 0;
	}

	cache->cleanup = cleanup;

	return cache;
}

void hash_cache_delete( struct hash_cache *cache )
{
	if(cache) {
		if(cache->table) hash_table_delete(cache->table);
		free(cache);
	}
}

int hash_cache_insert( struct hash_cache *cache, const char *key, void *value, int lifetime )
{
	struct entry *e, *old;

	e = malloc(sizeof(*e));
	if(!e) return 0;

	e->value = value;
	e->expires = time(0)+lifetime;

	hash_table_insert(cache->table,key,e,(void**)&old);
	if(old) {
		cache->cleanup(old->value);
		free(old);
	}

	return 1;
}

void * hash_cache_remove( struct hash_cache *cache, const char *key )
{
	struct entry *e;
	void *result;

	e = hash_table_remove(cache->table,key);
	if(e) {
		result = e->value;
		if(e->expires<time(0)) {
			cache->cleanup(result);
			result = 0;
		}
		free(e);
	} else {
		result = 0;
	}

	return result;
}

void * hash_cache_lookup( struct hash_cache *cache, const char *key )
{
	struct entry *e;
	void *result;

	e = hash_table_lookup(cache->table,key);
	if(e) {
		result = e->value;
		if(e->expires<time(0)) {
			result = hash_cache_remove(cache,key);
			cache->cleanup(result);
			result = 0;
		}
	} else {
		result = 0;
	}

	return result;
}

struct iteration_info {
	struct hash_cache *cache;
	hash_cache_iter_t iter;
	void *arg;
	time_t current;
};

static int iterate_if_fresh( struct hash_table *h, char *key, void *value, void *arg )
{
	struct entry *e = value;
	struct iteration_info *info = arg;

	if(e->expires<info->current) {
		hash_cache_remove(info->cache,key);
		return 1;
	} else {
		return info->iter(info->cache,key,e->value,info->arg);
	}
}

int hash_cache_iterate( struct hash_cache *h, hash_cache_iter_t iter, void *arg )
{
	struct iteration_info info;
	info.cache = h;
	info.iter = iter;
	info.arg = arg;
	info.current = time(0);
	return hash_table_iterate(h->table,iterate_if_fresh,&info);
}
