
#include "nvpair.h"
#include "hash_table.h"
#include "stringtools.h"
#include "xmalloc.h"

#include <stdlib.h>
#include <string.h>

struct nvpair {
	struct hash_table *table;
};

struct nvpair * nvpair_create()
{
	struct nvpair *n;
	n = xxmalloc(sizeof(*n));
	n->table = hash_table_create(127,hash_string);
	return n;
}

static int remove_item( struct hash_table *table, char *key, void *value, void *arg )
{
	hash_table_remove(table,key);
	free(value);
	return 1;
}

void nvpair_delete( struct nvpair *n )
{
	hash_table_iterate(n->table,remove_item,0);
	hash_table_delete(n->table);
	free(n);
}

void nvpair_parse( struct nvpair *n, const char *data )
{
	char *text = xstrdup(data);
	char *name, *value;

	name = strtok(text," ");
	while(name) {
		value = strtok(0,"\n");
		if(value) {
			nvpair_insert_string(n,name,value);
		} else {
			break;
		}
		name = strtok(0," ");
	}

	free(text);
}

void nvpair_insert_string( struct nvpair *n, const char *name, const char *value )
{
	void *old;
	hash_table_insert(n->table,name,xstrdup(value),&old);
	if(old) free(old);
}

void nvpair_insert_integer( struct nvpair *n, const char *name, INT64_T ivalue )
{
	char value[256];
	sprintf(value,"%lld",ivalue);
	nvpair_insert_string(n,name,(char*)value);
}

const char * nvpair_lookup_string( struct nvpair *n, const char *name )
{
	return hash_table_lookup(n->table,name);
}

INT64_T nvpair_lookup_integer( struct nvpair *n, const char *name )
{
	const char *value;
	value = hash_table_lookup(n->table,name);
	if(value) {
		return atoll(value);
	} else {
		return 0;
	}
}

static int print_text( struct hash_table *table, char *key, void *value, void *arg )
{
	FILE *stream = arg;
	fprintf(stream,"%s %s\n",key,(char*)value);
	return 1;
}

void nvpair_print_text( struct nvpair *n, FILE *stream )
{
	hash_table_iterate(n->table,print_text,stream);
	fprintf(stream,"\n");
}

static int print_xml( struct hash_table *table, char *key, void *value, void *arg )
{
	FILE *stream = arg;
	fprintf(stream,"<%s>%s</%s>\n",key,(char*)value,key);
	return 1;
}

void nvpair_print_xml( struct nvpair *n, FILE *stream )
{
	fprintf(stream,"<item>\n");
	hash_table_iterate(n->table,print_xml,stream);
	fprintf(stream,"</item>\n\n");
}

static int print_old_classads( struct hash_table *table, char *key, void *value, void *arg )
{
	FILE *stream = arg;
	fprintf(stream,"%s = \"%s\"\n",key,(char*)value);
	return 1;
}

void nvpair_print_old_classads( struct nvpair *n, FILE *stream )
{
	hash_table_iterate(n->table,print_old_classads,stream);
	fprintf(stream,"\n");
}

static int print_new_classads( struct hash_table *table, char *key, void *value, void *arg )
{
	FILE *stream = arg;
	fprintf(stream,"%s = \"%s\";\n",key,(char*)value);
	return 1;
}

void nvpair_print_new_classads( struct nvpair *n, FILE *stream )
{
	fprintf(stream,"[\n");
	hash_table_iterate(n->table,print_new_classads,stream);
	fprintf(stream,"]\n");
}

#define COLOR_ONE "#aaaaff"
#define COLOR_TWO "#bbbbbb"

static int color_counter=0;

void nvpair_print_html_header( FILE *stream, const char *headers[] )
{
	fprintf(stream,"<table bgcolor=%s>\n",COLOR_TWO);
	fprintf(stream,"<tr bgcolor=%s>\n",COLOR_ONE);
	while(*headers) {
		fprintf(stream,"<td align=right><b>%s</b>\n",*headers);
		headers++;
	}
	color_counter = 0;
}

void nvpair_print_html( struct nvpair *n, FILE *stream, const char *headers[] )
{
	fprintf(stream,"<tr bgcolor=%s>\n",color_counter%2 ? COLOR_ONE : COLOR_TWO );
	color_counter++;
	while(*headers) {
		const char *text = nvpair_lookup_string(n,*headers);
		if(!text) text = "???";
		fprintf(stream,"<td align=right>");
		if(!strcmp(*headers,"url")) {
			fprintf(stream,"<a href=%s>%s</a>\n",text,text);
		} else if(string_is_integer(text)) {
			char line[1024];
			string_metric(atof(text),-1,line);
			fprintf(stream,"%sB\n",line);
		} else {
			fprintf(stream,"%s\n",text);
		}
		headers++;
	}
}

void nvpair_print_html_footer( FILE *stream, const char *headers[] )
{
	fprintf(stream,"</table>\n");
}

