
#include "debug.h"
#include "getopt.h"
#include "stringtools.h"
#include "link.h"
#include "copy_stream.h"
#include "domain_name_cache.h"
#include "username.h"
#include "create_dir.h"
#include "copy_stream.h"
#include "full_io.h"

#include "gdbm.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <math.h>
#include <ctype.h>
#include <sys/stat.h>
#include <utime.h>

#define KEY_MAX 4096
#define MYLINE_MAX 4096
#define VALUE_MAX 65536

char * data_path = INSTALL_PATH "/etc";
double cut_xpos = -180;
int whois_timeout = 60;
int whois_retry=0;

FILE *summary_file=0;
FILE *country_file=0;
FILE *failure_file=0;

GDBM_FILE exception_db;
GDBM_FILE metawhois_db;
GDBM_FILE whois_db;
GDBM_FILE country_db;
GDBM_FILE city_db;
GDBM_FILE coord_db;

int total_count=0;
int success_exception_count=0;
int success_whois_count=0;
int success_country_count=0;
int failure_name_count=0;
int failure_whois_count=0;
int failure_location_count=0;

typedef enum {
	MAP_SUCCESS_EXCEPTION=1,
	MAP_SUCCESS_WHOIS=2,
	MAP_SUCCESS_COUNTRY=3,
	MAP_FAILURE_NAME=-1,
	MAP_FAILURE_WHOIS=-2,
	MAP_FAILURE_LOCATION=-3
} map_result_t;

typedef struct {
	double xpos;
	double ypos;
	char country[MYLINE_MAX];
	char state[MYLINE_MAX];
	char city[MYLINE_MAX];
} map_location_t;

void normalize_case( char *s )
{
	int gotspace=1;
	while(*s) {
		if(isspace(*s)) {
			gotspace=1;
		} else if(gotspace) {
			*s = toupper(*s);
			gotspace = 0;
		} else {
			*s = tolower(*s);
		}
		s++;
	}	
}

int mdb_firstkey( GDBM_FILE db, char *key )
{
	datum dbkey;

	dbkey = gdbm_firstkey(db);
	if(dbkey.dptr) {
		strcpy(key,dbkey.dptr);
		key[dbkey.dsize] = 0;
		free(dbkey.dptr);
		return 1;
	} else {
		return 0;
	}
}

int mdb_nextkey( GDBM_FILE db, const char *oldkey, char *key )
{
	datum olddbkey, dbkey;

	olddbkey.dptr = (char*) oldkey;
	olddbkey.dsize = strlen(oldkey);

	dbkey = gdbm_nextkey(db,olddbkey);
	if(dbkey.dptr) {
		strcpy(key,dbkey.dptr);
		key[dbkey.dsize] = 0;
		free(dbkey.dptr);
		return 1;
	} else {
		return 0;
	}
}

int mdb_lookup( GDBM_FILE db, const char *key, char *value )
{
	datum dbkey, dbvalue;
	dbkey.dptr=(char*)key;
	dbkey.dsize=strlen(key);
	dbvalue = gdbm_fetch(db,dbkey);
	if(dbvalue.dptr) {
		strcpy(value,dbvalue.dptr);
		value[dbvalue.dsize] = 0;
		return 1;
	} else {
		return 0;
	}	
}

int mdb_lookup_longest_suffix( GDBM_FILE db, const char *key, char *value )
{
	char *d;

	value[0] = 0;

	d = strrchr(key,'.');
	while(d) {
		if(mdb_lookup(db,d,value)) {
			do {
				d--;
			} while(d>=key && *d!='.');
			continue;
		} else {
			break;
		}
	}

	if(value[0]) {
		return 1;
	} else {
		return 0;
	}
}

int mdb_store( GDBM_FILE db, const char *key, const char *value )
{
	datum dbkey, dbvalue;
	dbkey.dptr=(char*)key;
	dbkey.dsize=strlen(key);
	dbvalue.dptr=(char*)value;
	dbvalue.dsize=strlen(value);
	return gdbm_store(db,dbkey,dbvalue,GDBM_REPLACE);
}

GDBM_FILE mdb_load( const char *basename )
{
	char loadfile[PATH_MAX];
	char dbfile[PATH_MAX];
	struct stat loadstat;
	struct stat dbstat;
	FILE *file;
	GDBM_FILE db;
	int do_load = 0;
	int result;

	sprintf(loadfile,"%s/mapaddr.%s.txt",data_path,basename);
	sprintf(dbfile,"%s/mapaddr.%s.db",data_path,basename);

	file = fopen(loadfile,"r");
	if(!file) fatal("couldn't open %s: %s",loadfile,strerror(errno));
	fstat(fileno(file),&loadstat);

	result = stat(dbfile,&dbstat);
	if(result!=0 || (dbstat.st_mtime<loadstat.st_mtime)) {
		do_load = 1;
	} else {
		do_load = 0;
	}

	db = gdbm_open(dbfile,0,GDBM_WRCREAT,0777,0);
	if(!db) fatal("couldn't open %s: %s",dbfile,strerror(errno));

	if(do_load) {
		char *key, *value;
		char line[1024];
		debug(D_NOTICE,"updating %s database",basename);
		while(fgets(line,sizeof(line),file)) {
			string_chomp(line);
			if(line[0]=='#') continue;
			key = line;
			value = strchr(line,':');
			if(!value) continue;
			*value = 0;
			value++;
			mdb_store(db,key,value);
		}
		utime(dbfile,0);
	}

	return db;
}

int whois_lookup( const char *host, const char *key, char *value, time_t stoptime )
{
	char addr[DOMAIN_NAME_MAX];
	char query[DOMAIN_NAME_MAX];
	struct link *link;
	int result;

	if(!domain_name_cache_lookup(host,addr)) return 0;

	link = link_connect(addr,43,stoptime);
	if(!link) return 0;

	sprintf(query,"%s\n",key);

	result = link_write(link,query,strlen(query),stoptime);
	if(result!=strlen(query)) {
		link_close(link);
		return 0;
	}

	result = link_read(link,value,VALUE_MAX,stoptime);
	link_close(link);

	if(result>0) {
		value[result] = 0;
		return 1;
	} else {
		return 0;
	}
}

int whois_lookup_cached( const char *host, const char *key, char *value, time_t stoptime, int usecache )
{
	char dbkey[KEY_MAX];

	sprintf(dbkey,"%s@%s",key,host);

	if(usecache && mdb_lookup(whois_db,dbkey,value)) {
		debug(D_DEBUG,"cached whois query: %s",dbkey);
		return 1;
	}

	debug(D_DEBUG,"whois query: %s",dbkey);

	if(whois_lookup(host,key,value,stoptime)) {
		debug(D_DEBUG,"whois query ok");
		mdb_store(whois_db,dbkey,value);
		return 1;
	} else {
		debug(D_DEBUG,"whois query failed");
		return 0;
	}
}


int match_country( const char *value, map_location_t *location )
{
	char line[MYLINE_MAX];
	char country_code[MYLINE_MAX];
	char country_name[MYLINE_MAX];

	string_toupper((char*)value);

	debug(D_DEBUG,"searching for country name in result");
	fseek(country_file,0,SEEK_SET);
	while(fgets(line,sizeof(line),country_file)) {
		if(line[0]=='#') continue;
		if(line[0]=='\n') continue;
		if(sscanf(line,"%[^:]:%[^:]:",country_code,country_name)==2) {
			string_toupper(country_code);
			string_toupper(country_name);
			if(strstr(value,country_name)) {
				debug(D_DEBUG,"found country %s (%s) in result",country_name,country_code);
				strcpy(location->country,country_code);
				return 1;
			}
		}
	}
	debug(D_DEBUG,"no country name found");
	return 0;
}

int match_address( char *buffer, map_location_t *location )
{
	char test_city[1024];
	char test_state[1024];
	int test_zip;

	char *t = buffer;
	int gotcity = 0;
	int gotstate = 0;

	debug(D_DEBUG,"searching for U.S. address in result");

	while(t) {
		if(sscanf(t," %[a-zA-Z ], %[A-Z] %d",test_city,test_state,&test_zip)==3) {
			if(strlen(test_state)==2) {
				debug(D_DEBUG,"matched full street address: %s %s %d",test_city,test_state,test_zip);
				gotcity = 1;
				gotstate = 1;
				strcpy(location->city,test_city);
				strcpy(location->state,test_state);
			}
		} else if(sscanf(t," Registrant City:%[^\n\r]",test_city)) {
			string_toupper(test_city);
			debug(D_DEBUG,"matched city field: %s", test_city);
			gotcity = 1;
			strcpy(location->city,test_city);
		} else if(sscanf(t," Registrant State/Province:%[^\n\r]",test_state)) {
			if(strlen(test_state)==2) {
				string_toupper(test_state);
				debug(D_DEBUG,"matched state field: %s",test_state);
				gotstate = 1;
				strcpy(location->state,test_state);
			}
		}
		if(gotcity && gotstate) {
			strcpy(location->country,"US");
			return 1;
		}
		t = strchr(t+1,'\n');

	}
	return 0;
}

int whois_lookup_recursive( const char *starthost, const char *key, map_location_t *location, time_t stoptime, int usecache )
{
	char server[DOMAIN_NAME_MAX];
	char newserver[DOMAIN_NAME_MAX];
	char newkey[DOMAIN_NAME_MAX];
	char value[VALUE_MAX];
	int doequals;

	strcpy(server,starthost);

	retry:

	for(doequals=0;doequals<2;doequals++) {
		if(doequals) {
			sprintf(newkey,"=%s",key);
		} else {
			strcpy(newkey,key);
		}
		newserver[0] = 0;
		if(whois_lookup_cached(server,newkey,value,stoptime,usecache)) {
			char *n, *m;
			n = strstr(value,"Whois Server:");
			while(n) {
				m = strstr(n+1,"Whois Server:");
				if(m) {
					n = m;
					continue;
				} else {
					break;
				}
			}
			if(	n
				&& sscanf(n,"Whois Server: %s",newserver)
				&& strcmp(server,newserver)
			) {

				strcpy(server,newserver);
				goto retry;
			}
			if(match_address(value,location)) {
				return 1;
			} else if(match_country(value,location)) {
				return 1;
			} else {
				if(failure_file) {
					fprintf(failure_file,"whois %s@%s\n",newkey,server);
					fprintf(failure_file,"%s\n",value);
				}
			}
		}
	}

	return 0;
}			

int whois_search( const char *key, map_location_t *location, time_t stoptime, int usecache )
{
	char host[DOMAIN_NAME_MAX];

	if(!mdb_lookup_longest_suffix(metawhois_db,key,host)) {
		debug(D_DEBUG,"couldn't find whois server for %s",key);
		return 0;
	}

	debug(D_DEBUG,"whois server for %s is %s",key,host);

	while(!whois_lookup_recursive(host,key,location,stoptime,usecache)) {
		key = strchr(key,'.');
		if(key) {
			key++;
			if(strchr(key,'.')) {
				continue;
			} else {
				return 0;
			}
		} else {
			return 0;
		}
	}
	return 1;
}

int whois_search_retry( const char *key, map_location_t *location, time_t stoptime )
{
	if(whois_search(key,location,stoptime,1)) {
		return 1;
	} else if(whois_retry) {
		return whois_search(key,location,stoptime,0);
	} else {
		return 0;
	}
}

int exception_search( const char *name, map_location_t *location )
{
	char value[VALUE_MAX];

	if(mdb_lookup_longest_suffix(exception_db,name,value)) {
		sscanf(value,"%[^:]:%[^:]:%[^:]",location->country,location->city,location->state);
		debug(D_DEBUG,"found %s in exception table",name);
		return 1;
	} else {
		debug(D_DEBUG,"didn't find %s in exception table",name);
		return 0;
	}
}

int country_search( const char *name, map_location_t *location )
{
	if(name[strlen(name)-3]=='.') {
		char country[KEY_MAX];
		char value[VALUE_MAX];
		strcpy(country,&name[strlen(name)-2]);
		string_toupper(country);
		if(mdb_lookup(country_db,country,value)) {
			strcpy(location->country,country);
			strcpy(location->city,"UNKNOWN");
			strcpy(location->state,"XX");
			debug(D_DEBUG,"found %s by country suffix",name);
			return 1;
		}
	}
	debug(D_DEBUG,"%s does not have a valid country suffix",name);
	return 0;
}

int city_locate( map_location_t *location )
{
	char key[KEY_MAX];
	char value[VALUE_MAX];
	int dn,mn,sn,dw,mw,sw;
	char ns,ew;

	sprintf(key,"%s/%s",location->city,location->state);
	string_toupper(key);

	if(mdb_lookup(city_db,key,value)) {
		sscanf(value,"%d:%d:%d:%c:%d:%d:%d:%c",&dw,&mw,&sw,&ew,&dn,&mn,&sn,&ns);
		location->ypos = dn+mn/60.0+sn/3600.0;
		location->xpos = dw+mw/60.0+sw/3600.0;
		if(ns!='N') location->ypos = -location->ypos;
		if(ew!='E') location->xpos = -location->xpos;
		debug(D_DEBUG,"found %s, %s in city table",location->city,location->state);
		return 1;
	} else {
		debug(D_DEBUG,"not found in city table");
		return 0;
	}
}

int country_locate( map_location_t *location )
{
	char country_name[MYLINE_MAX];
	char country_capital[MYLINE_MAX];
	char value[VALUE_MAX];
	int dn,mn,sn,dw,mw,sw;
	char ns,ew;

	string_toupper(location->country);

	if(mdb_lookup(country_db,location->country,value)) {
		sscanf(value,"%[^:]:%[^:]:%d:%d:%d:%c:%d:%d:%d:%c",country_name,country_capital,&dw,&mw,&sw,&ew,&dn,&mn,&sn,&ns);
		location->ypos = dn+mn/60.0+sn/3600.0;
		location->xpos = dw+mw/60.0+sw/3600.0;
		if(ns!='N') location->ypos = -location->ypos;
		if(ew!='E') location->xpos = -location->xpos;
		debug(D_DEBUG,"found %s in country table",location->country);
		return 1;
	} else {
		debug(D_DEBUG,"%s not found in country table",location->country);
		return 0;
	}
}

char map_result_char( map_result_t result ) {
	total_count++;
	switch(result) {
		case MAP_SUCCESS_EXCEPTION:
			success_exception_count++;
			return 'e';
		case MAP_SUCCESS_WHOIS:
			success_whois_count++;
			return '.';
		case MAP_SUCCESS_COUNTRY:
			success_country_count++;
			return 'c';
		case MAP_FAILURE_NAME:
			failure_name_count++;
			return 'N';
		case MAP_FAILURE_WHOIS:
			failure_whois_count++;
			return 'W';
		case MAP_FAILURE_LOCATION:
			failure_location_count++;
			return 'L';
		default:
			return 'X';
	}
}

map_result_t map_one_name( const char *name )
{
	map_result_t result;
	map_location_t location;

	string_tolower((char*)name);

	strcpy(location.country,"XX");
	strcpy(location.state,"XX");
	strcpy(location.city,"Unknown");
	location.xpos = 240;
	location.ypos = -40;

	if(exception_search(name,&location)) {
		result = MAP_SUCCESS_EXCEPTION;
	} else if(!strchr(name,'.')) {
		debug(D_DEBUG,"%s is not a valid domain name",name);
		result = MAP_FAILURE_NAME;
	} else if(whois_search_retry(name,&location,time(0)+whois_timeout)) {
		result = MAP_SUCCESS_WHOIS;
	} else if(country_search(name,&location)) {
		result = MAP_SUCCESS_COUNTRY;
	} else {
		debug(D_DEBUG,"unable to locate %s",name);
		result = MAP_FAILURE_WHOIS;
	}

	if(result>0) {
		if(!strcmp(location.country,"US")) {
			if(!city_locate(&location)) {
				debug(D_NOTICE,"unknown city/state %s/%s",location.city,location.state);
				result = MAP_FAILURE_LOCATION;
			}
		} else {
			if(!country_locate(&location)) {
				debug(D_NOTICE,"unknown country code %s",location.country);
				result = MAP_FAILURE_LOCATION;
			}
		}
	}

	while(location.xpos<cut_xpos) {
		location.xpos += 360;
	}
	while(location.xpos>(cut_xpos+360)) {
		location.xpos -= 360;
	}
	normalize_case(location.city);
	
	printf("%lf:%lf:%s:%s:%s\n",location.xpos,location.ypos,location.country,location.state,location.city);

	return result;
}

static void show_use( const char *cmd ) 
{
        fprintf(stderr,"Use: %s [options] <name>\n",cmd);
        fprintf(stderr,"Where options are:\n");
	fprintf(stderr," -c <deg>    Cut longitude here. (default=%5f)\n",cut_xpos);
        fprintf(stderr," -d <flags>  Debugging flags\n");
	fprintf(stderr," -D <path>   Location of data files (default is %s)\n",data_path);
	fprintf(stderr," -f <file>   Send failed whois searches to this file\n");
	fprintf(stderr," -o <file>   Send debug messages to this file\n");
	fprintf(stderr," -s <file>   Send summary statistics to this file\n");
	fprintf(stderr," -t <secs>   Timeout on whois queries\n");
	fprintf(stderr," -q          Quiet mode.\n");
	fprintf(stderr," -r          Retry failed whois lookups\n");
	fprintf(stderr," -v          Show version\n");
	fprintf(stderr," -h          Show this help\n");
        fprintf(stderr,"Where debug flags arg: ");
        debug_flags_print(stderr);
        fprintf(stderr,"\n");
	fprintf(stderr,"If <name> is - then a series of names will be read from standard input\n");
}

static void show_version( char *cmd )
{
        fprintf(stderr,"mapaddr version %d.%d.%d built by %s@%s at %s on %s\n",CCTOOLS_VERSION_MAJOR,CCTOOLS_VERSION_MINOR,CCTOOLS_VERSION_MICRO,BUILD_USER,BUILD_HOST,__TIME__,__DATE__);
}

int main( int argc, char *argv[] )
{
	char country_path[PATH_MAX];
	char *name;
	char c;

	debug_config(argv[0]);

        while((c=getopt(argc,argv,"d:o:D:r:c:s:f:t:qrvh"))!=(char)-1) {
		switch(c) {
			case 'd':
				debug_flags_set(optarg);
				break;
			case 'o':
				debug_config_file(optarg);
				break;
			case 'D':
				data_path = optarg;
				break;
			case 'c':
				cut_xpos = atof(optarg);
				break;
			case 's':
				summary_file = fopen(optarg,"w");
				if(!summary_file) fatal("couldn't open %s: %s",optarg,strerror(errno));
				break;
			case 'f':
				failure_file = fopen(optarg,"w");
				if(!failure_file) fatal("couldn't open %s: %s",optarg,strerror(errno));
				break;
			case 't':
				whois_timeout = atoi(optarg);
				break;
			case 'q':
				debug_flags_restore(0);
				break;
			case 'r':
				whois_retry = 1;
				break;
			case 'v':
				show_version(argv[0]);
				exit(1);
				break;
			default:
				show_use(argv[0]);
				exit(1);
				break;
		}
	}

	if(argc==optind) {
		show_use(argv[0]);
		exit(1);
	}

	name = argv[optind];

	exception_db = mdb_load("exception");
	metawhois_db = mdb_load("metawhois");
	whois_db = mdb_load("whois");
	country_db = mdb_load("country");
	city_db = mdb_load("city");
	coord_db = mdb_load("coord");

	sprintf(country_path,"%s/mapaddr.country.txt",data_path);

	country_file = fopen(country_path,"r");
	if(!country_file) fatal("couldn't open %s: %s",country_path,strerror(errno));

	srand(time(0)+getpid());

	if(!strcmp(name,"-")) {
		char line[4096];
		map_result_t result;
		while(fgets(line,sizeof(line),stdin)) {
			string_chomp(line);
			name = strrchr(line,':');
			if(name) {
				name++;
			} else {
				name=line;
			}
			printf("%s:",line);
			result = map_one_name(name);
			fprintf(stderr,"%c",map_result_char(result));
		}
		fprintf(stderr,"\n");
	} else {
		printf("%s:",name);
		map_one_name(name);
	}

	if(summary_file) {
		fprintf(summary_file,"total-all:         %d\n",total_count);
		fprintf(summary_file,"success-all:       %d\n",success_exception_count+success_whois_count+success_country_count);
		fprintf(summary_file,"success-whois:     %d\n",success_whois_count);
		fprintf(summary_file,"success-country:   %d\n",success_country_count);
		fprintf(summary_file,"success-exception: %d\n",success_exception_count);
		fprintf(summary_file,"failure-all:       %d\n",failure_name_count+failure_whois_count+failure_location_count);
		fprintf(summary_file,"failure-name:      %d\n",failure_name_count);
		fprintf(summary_file,"failure-whois      %d\n",failure_whois_count);
		fprintf(summary_file,"failure-location   %d\n",failure_location_count);
		fclose(summary_file);
	}

	fflush(0);

	return 0;
}

